
from typing import Any, ClassVar, Optional

from cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject import (
    ReadResponse,
    SchemaResponse,
    Summary,
)
from cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Core import ServiceInfoResponse
from cherwell_pydantic_api.api import Connection
from cherwell_pydantic_api.bo.registry import BusinessObjectRegistry
from cherwell_pydantic_api.interfaces import ApiRequesterInterface
from cherwell_pydantic_api.settings import InstanceSettingsBase, Settings
from cherwell_pydantic_api.types import BusinessObjectType, BusObID, BusObRecID
from cherwell_pydantic_api.utils import docwraps



class Instance(ApiRequesterInterface):
    _instances: ClassVar[dict[str, "Instance"]] = {}
    __call_token: ClassVar[object] = object()

    def __init__(self, instance_settings: InstanceSettingsBase, __called_from_use: Any = 'Please create Instance objects using Instance.use()'):
        if __called_from_use is not self.__call_token:
            raise TypeError(
                'Please create Instance objects using Instance.use()')
        # Take into account that the settings might change under our nose
        self._settings = instance_settings.copy()
        self.bo = BusinessObjectRegistry(self)
        self._connection = Connection(instance_settings)


    @property
    def settings(self) -> InstanceSettingsBase:
        return self._settings


    @docwraps(Connection.authenticate)
    async def authenticate(self):
        return await self._connection.authenticate()


    @docwraps(Connection.logout)
    async def logout(self):
        return await self._connection.logout()


    @docwraps(Connection.get_busobid)
    async def get_busobid(self, busobname: str) -> Optional[BusObID]:
        return await self._connection.get_busobid(busobname)


    @docwraps(Connection.GetBusinessObjectSchemaV1)
    async def get_bo_schema(self, *, busobid: Optional[BusObID] = None, busobname: Optional[str] = None) -> Optional[SchemaResponse]:
        if busobid is None:
            if busobname is None:
                raise TypeError(
                    'get_bo_schema() missing 1 required argument: busobid or busobname')
            busobid = await self.get_busobid(busobname)
            if busobid is None:
                return None
        response = await self._connection.GetBusinessObjectSchemaV1(busobId=busobid, includerelationships=True)
        if not response.hasError:
            self.bo.register(response)
        return response


    @docwraps(Connection.get_bo_summaries)
    async def get_bo_summaries(self, type: BusinessObjectType = "Major") -> list[Summary]:
        response = await self._connection.get_bo_summaries(type)
        for summary in response:
            self.bo.register_summary(summary)
        return response


    @docwraps(Connection.GetBusinessObjectByRecIdV1)
    async def get_bo(self, busobid: BusObID, *, busobrecid: Optional[BusObRecID] = None, publicid: Optional[str] = None) -> ReadResponse:
        if busobrecid is None:
            if publicid is None:
                raise TypeError(
                    'get_bo() missing 1 required argument: busobrecid or publicid')
            return await self._connection.GetBusinessObjectByPublicIdV1(busobid=busobid, publicid=publicid)
        return await self._connection.GetBusinessObjectByRecIdV1(busobid=busobid, busobrecid=busobrecid)


    @docwraps(Connection.GetServiceInfoV1)
    async def get_service_info(self) -> ServiceInfoResponse:
        response = await self._connection.GetServiceInfoV1()
        self.bo.register_service_info(response)
        return response


    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}({self.settings.name})>"


    @classmethod
    def use(cls, name: Optional[str] = None, *, instance_settings: Optional[InstanceSettingsBase] = None) -> "Instance":
        """Instantiate an Instance object for the given instance name, or return an existing Instance object if one already exists."""
        settings = Settings()
        if instance_settings is None:
            instance_settings = settings.get_instance(name)
        if instance_settings is None:
            raise ValueError(f"Instance {name} not found in settings")
        name = instance_settings.name
        if name not in cls._instances:
            cls._instances[name] = cls(instance_settings, cls.__call_token)
        return cls._instances[name]



import types
from functools import wraps
from typing import Any, Callable, Coroutine, Optional, TypeVar

from cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject import (
    ReadResponse,
    Summary,
)
from cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Core import ServiceInfoResponse
from cherwell_pydantic_api.bo.registry import BusinessObjectRegistry
from cherwell_pydantic_api.bo.wrapper import BusinessObjectWrapper
from cherwell_pydantic_api.instance import Instance
from cherwell_pydantic_api.types import BusinessObjectType



__all__ = ['Interactive']


_ReturnType = TypeVar("_ReturnType")
_WaiterType = Callable[[Coroutine[Any, Any, _ReturnType]], _ReturnType]


class RestaurantInterface:
    "A restaurant contains a waiter."
    _waiter: _WaiterType


def wait(amethod: Callable[..., Coroutine[Any, Any, _ReturnType]]) -> Callable[..., _ReturnType]:
    """Decorator to convert an async method into a sync method, using the waiter method of the Interactive instance.
    """

    @wraps(amethod)
    def wrapper(self: RestaurantInterface, *args, **kwargs):
        return self._waiter(amethod(self, *args, **kwargs))
    return wrapper


class WaiterProxy(RestaurantInterface):
    """A proxy around an object with async methods that uses the waiter method to make all public methods sync."""

    def __init__(self, async_obj: object, waiter: _WaiterType):
        self._async_obj = async_obj
        self._waiter = waiter
        self.__doc__ = async_obj.__doc__
        self.__wrapped__ = async_obj

    def __getattr__(self, name: str) -> Any:
        if name.startswith('_'):
            raise AttributeError(name)
        amethod = getattr(self._async_obj, name)
        if not callable(amethod):
            return amethod

        @wraps(amethod)
        def wrapper(*args, **kwargs):
            r = amethod(*args, **kwargs)
            if isinstance(r, types.CoroutineType):
                return self._waiter(r)
            return r
        return wrapper

    def __setattr__(self, name: str, value) -> None:
        if not name.startswith('_') and hasattr(self._async_obj, name):
            setattr(self._async_obj, name, value)
        else:
            super().__setattr__(name, value)

    def __delattr__(self, name: str) -> None:
        if not name.startswith('_') and hasattr(self._async_obj, name):
            delattr(self._async_obj, name)
        else:
            super().__delattr__(name)

    def __dir__(self) -> list[str]:
        return dir(self._async_obj)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self._async_obj!r})"


@wraps(BusinessObjectWrapper)
def create_bo_wrapper_class(waiter: _WaiterType) -> type[BusinessObjectWrapper]:
    class InteractiveBusinessObjectWrapper(BusinessObjectWrapper, RestaurantInterface):
        _waiter = waiter

        @wait
        async def get(self, publicid: str) -> ReadResponse:
            return await super().get(publicid=publicid)

    return InteractiveBusinessObjectWrapper



class Interactive(RestaurantInterface):
    """A wrapper around various cherwell_pydantic_api objects, using the waiter method to make all methods sync (no await needed).
    Each instance of Interactive wraps a specific Instance and indirectly the Connection object.
    It also wraps the BusinessObjectRegistry, so that all BusinessObjectWrapper objects created by it can be accessed with using await.
    """

    def __init__(self, *, instance_name: Optional[str] = None, waiter: Optional[_WaiterType] = None):
        self._instance = Instance.use(name=instance_name)
        if waiter is None:
            import asyncio
            waiter = asyncio.run
        self._waiter = waiter
        self._instance_proxy = WaiterProxy(self._instance, self._waiter)
        self._connection_proxy = WaiterProxy(
            self._instance._connection, self._waiter)
        self.help: Any = ''

    @property
    def connection(self) -> WaiterProxy:
        "WaiterProxy wrapper around the Connection object."
        return self._connection_proxy

    @property
    def instance(self) -> WaiterProxy:
        "WaiterProxy wrapper around the Instance object."
        return self._instance_proxy

    def async_wrap(self, **kwargs) -> "Interactive":
        "Wrap the given objects with WaiterProxy and add them as attributes to the Interactive instance."
        # TODO: restrict attribute name
        for attr, obj in kwargs.items():
            setattr(self, attr, WaiterProxy(obj, self._waiter))
        return self

    @wait
    async def authenticate(self):
        await self._instance.authenticate()

    @wait
    async def get_bo_schema(self, busobname: str):
        return await self._instance.get_bo_schema(busobname=busobname)

    @wait
    async def get_bo_summaries(self, type: BusinessObjectType = "All") -> list[Summary]:
        return await self._instance.get_bo_summaries(type=type)

    @wait
    async def get_service_info(self) -> ServiceInfoResponse:
        return await self._instance.get_service_info()

    @property
    def bo(self) -> BusinessObjectRegistry:
        self._instance.bo.set_wrapper_class(
            create_bo_wrapper_class(self._waiter))
        return self._instance.bo

    def __repr__(self):
        return f"{self.__class__.__name__}(instance={self._instance!r})"



#!/usr/bin/env python3

from typing import TYPE_CHECKING, Optional

import click



@click.command('ipython', context_settings={'allow_extra_args': True, 'ignore_unknown_options': True}, epilog="Further arguments will be passed to IPython.")
@click.option('--instance-name', '-I', help='The name of the instance to use. If not specified, the default instance will be used.')
@click.option('--quiet', '-q', is_flag=True, help='Suppress most messages')
@click.pass_context
def ipython_shell(ctx, instance_name: Optional[str] = None, quiet: bool = False):
    "Start an interactive IPython shell"
    try:
        import IPython
        import traitlets
    except:
        if not quiet:
            click.secho("IPython is not installed", fg='red')
        return

    from cherwell_pydantic_api.bo.modelgen.collector import Collector
    from cherwell_pydantic_api.bo.modelgen.repo import ModelRepo
    from cherwell_pydantic_api.instance import Instance
    from cherwell_pydantic_api.interactive import Interactive
    from cherwell_pydantic_api.settings import Settings

    collector = Collector(Instance.use(instance_name
                                       ), verbose=True, bo_include_filter=r'(?i)ticket|task|changerequest|conf.*|journal.*|spec.*|.*port*')
    repo = ModelRepo(create=True)

    def startup(objs):
        cw = Interactive(instance_name=instance_name,
                         waiter=IPython.get_ipython().loop_runner)  # type: ignore
        if not quiet:
            click.secho('\nWelcome to cherwell_pydantic_api!', fg='cyan')
            click.secho(
                f'Connecting to instance "{cw.instance.settings.name}" at {cw.instance.settings.base_url}...', nl=False, fg='magenta')
        cw.authenticate()
        service_info = cw.get_service_info()
        if not quiet:
            click.secho(' OK', fg='green')
            click.secho(
                f'apiVersion={service_info.apiVersion}, csmVersion={service_info.csmVersion}', fg='cyan')
            click.secho('I have created a ' + click.style('cw', fg='bright_cyan') +
                        ' object for you to use. Type ' + click.style('cw.help', fg='bright_cyan') + ' for details.')
            click.secho('Other useful globals: ' + ', '.join(objs))
        cw.async_wrap(collector=collector)

        class help:
            msg = ["\nGetting started:",
                   "> " + click.style("cw.collector.collect()", fg='cyan') + " - collect some business object IDs; this fills cw.bo",
                   "> " + click.style("[bo for bo in cw.bo.keys()]", fg='cyan') + " - list the business objects collected",
                   "Suppose this list includes 'ticket'. Then:",
                   "> " + click.style("cw.bo.ticket.get('20220707-1452')", fg='cyan') + " - get a ticket by its public ID",
                   ]

            def _ipython_display_(self):
                click.secho('\n'.join(self.msg))

        cw.help = help()
        return cw

    # Create a namespace for the shell
    ns = {}
    for name, obj in locals().items():
        if not (name.startswith('_') or name in ('ctx', 'IPython', 'traitlets', 'quiet', 'ns')):
            ns[name] = obj
    ns['click'] = click
    ns['object_help'] = [k for k in ns.keys() if k not in ('object_help', 'startup')]
    ns['object_help'].sort()
    config = traitlets.config.Config()  # type: ignore
    config.InteractiveShellApp.exec_lines = [
        "cw = startup(object_help)",
        "del startup",
        "del object_help",
    ]
    if quiet:
        config.InteractiveShellApp.display_banner = False
    IPython.start_ipython(config=config, user_ns=ns, argv=ctx.args)



if __name__ == '__main__':
    ipython_shell()



from cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject import ReadResponse
from cherwell_pydantic_api.bo.valid_schema import ValidSchema
from cherwell_pydantic_api.interfaces import ApiRequesterInterface
from cherwell_pydantic_api.types import BusObID, BusObIdentifier


class BusinessObjectWrapperBase:
    """Base class for BusinessObjectWrapper. A subclass is passed to the BusinessObjectRegistry constructor."""

    def __init__(self, *, schema: ValidSchema, instance: ApiRequesterInterface):
        pass


class BusinessObjectWrapper(BusinessObjectWrapperBase):
    """Wrap the Cherwell API SchemaResponse object to provide a more convenient interface to the business object schema.
    It is generally created by the BusinessObjectRegistry class.
    This is mainly intended for exploratory use e.g. in a Jupyter notebook or IPython.
    """

    def __init__(self, *, schema: ValidSchema, instance: ApiRequesterInterface):
        self._schema = schema
        self._instance = instance

    async def get(self, publicid: str) -> ReadResponse:
        "Get a business object by public ID."
        return await self._instance.get_bo(self._schema.busObId, publicid=publicid)

    @property
    def busobid(self) -> BusObID:
        "The business object ID."
        return self._schema.busObId

    @property
    def name(self) -> str:
        "The business object name, as reported by Cherwell."
        return self._schema.name

    @property
    def identifier(self) -> BusObIdentifier:
        "The business object identifier used by cherwell-pydantic-api. This will be a valid Python identifier."
        return self._schema.identifier

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} (BO type: {self._schema.name}, instance: {self._instance}, BusObId: {self._schema.busObId})>"



from typing import ClassVar, Optional, Type, TypeVar, cast

from pydantic import BaseModel

from cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject import ReadResponse
from cherwell_pydantic_api.bo.registry import BusinessObjectModelRegistryMixin
from cherwell_pydantic_api.instance import Instance
from cherwell_pydantic_api.types import BusObID, BusObRecID, CherwellAPIError, FieldID



BusObModel_T = TypeVar("BusObModel_T", bound="BusinessObjectModelBase")


class BusinessObjectModelBase(BaseModel, BusinessObjectModelRegistryMixin):
    _instance_name: ClassVar[str]

    def __init_subclass__(cls) -> None:
        # Check that _ModelData and _ModelData.busobid exist - this is not the case for intermediate classes
        if hasattr(cls, "_ModelData") and hasattr(cls._ModelData, "busobid"):
            # Register model classes in the instance's business object registry.
            cls.get_instance().bo.register_model(cls._ModelData.busobid, cls)
        return super().__init_subclass__()


    class _ModelData:
        busobid: BusObID
        statefieldid: Optional[FieldID]
        firstrecidfield: Optional[FieldID]
        recidfields: Optional[FieldID]


    @classmethod
    def get_instance(cls) -> Instance:
        """Get the instance object that this model is associated with."""
        return Instance.use(cls._instance_name)


    @classmethod
    def from_api_response(cls: Type[BusObModel_T], response: ReadResponse) -> BusObModel_T:
        assert response.fields is not None
        assert response.busObId == cls._ModelData.busobid
        assert all([f.name is not None for f in response.fields])
        vals = {cast(str, f.name): f.value for f in response.fields}
        return cls(**vals)


    @classmethod
    async def from_api(cls: Type[BusObModel_T], publicid: Optional[str] = None, *, busobrecid: Optional[BusObRecID] = None) -> Optional[BusObModel_T]:
        response = await cls.get_instance().get_bo(busobid=cls._ModelData.busobid, publicid=publicid, busobrecid=busobrecid)
        if response is None:
            raise ValueError("No response from Cherwell API")
        if response.hasError:
            raise CherwellAPIError(f"get_bo[{cls.__name__}]", errorMessage=response.errorMessage,
                                   errorCode=response.errorCode, httpStatusCode=response.httpStatusCode)
        return cls.from_api_response(response)



"""Module for the welcoming initial setup CLI"""

import time
from pathlib import Path

import click
import httpx
try:
    import readline
except:
    pass


def api_detect(urls):
    "Try to detect the Cherwell API at the given URLs"
    for url in urls:
        try:
            click.secho(
                f"Trying to detect Cherwell API at {url}...", fg='cyan')
            t1 = time.time()
            verify = True
            try:
                response = httpx.get(
                    f"{url}/api/V1/serviceinfo", verify=verify, timeout=(5.0, 60.0, 5.0, 1.0))
            except httpx.ConnectError:
                verify = False
                response = httpx.get(
                    f"{url}/api/V1/serviceinfo", verify=verify, timeout=(5.0, 60.0, 5.0, 1.0))
            t2 = time.time()
            if response.is_success:
                service_info = response.json()
                return (url, service_info.get('apiVersion', 'unknown'), t2 - t1, verify)
            click.secho(
                f"HTTP Error: {response.status_code} {response.reason_phrase}", fg='red')
        except Exception as e:
            click.secho(f"Error: {e}", fg='red')
    return (None, None, 0.0, False)


def initial_setup(envpath: Path):
    "Create a new cherwell.env file from user input"
    envdict = {}
    click.secho("""Enter the API base URL of your Cherwell instance, e.g. https://myinstance.cherwellsoftware.com/CherwellAPI
You can also just enter the hostname or IP address and I will attempt to guess the rest.
If you have multiple instances, set up the development instance first. You can add your production instance later.\n""", fg='green')
    while True:
        api_base_url = click.prompt('API base URL or host', type=str)
        if not api_base_url:
            continue
        if not api_base_url.startswith('http'):
            urls = [f"https://{api_base_url}/CherwellAPI",
                    f"http://{api_base_url}/CherwellAPI"]
        else:
            urls = [api_base_url]
        api_base_url, api_version, duration, verify = api_detect(urls)
        if api_base_url is not None:
            break
        click.secho(
            f"No API found. Please check your URL and network connectivity.", fg='red')
    click.secho(
        f"Success! Cherwell API version {api_version} found at {api_base_url}", bg='green', fg='black')
    envdict['cherwell_base_url'] = api_base_url
    duration = round(duration, 1)
    if duration > 5.0:
        click.secho(f"Warning: The API took {duration} seconds to respond. I will configure a timeout of {duration*2} seconds.", fg='yellow')
        envdict['cherwell_timeout'] = duration * 2
    if not verify:
        click.secho(
            f"Warning: SSL certificate verification failed. Connection will be set up with verify=False", fg='red', bg='yellow')
        envdict['cherwell_verify'] = 'off'
    click.echo(f"""
You can now enter your client ID, username and password. You can also leave these blank and configure them later,
however you won't be able to use the API until you do. Contact your Cherwell administrator if you don't have these details.
""")
    envdict['cherwell_client_id'] = click.prompt(
        'Client ID', type=str, default='')
    envdict['cherwell_username'] = click.prompt(
        'Username', type=str, default='')
    envdict['cherwell_password'] = click.prompt(
        'Password', type=str, default='')

    click.echo('Would you like to create the file ', nl=False)
    click.secho(f"{envpath}", fg='cyan', nl=False)
    click.echo(' with the above settings?')
    if click.confirm('Create file?', default=None):
        envpath.write_text(''.join([f"{k}={v}\n" for k, v in envdict.items()]))
        click.secho(f"{envpath} created successfully:", fg='green')





import re
from typing import TYPE_CHECKING, Any, AsyncIterable, Iterable, Literal, Mapping, Optional, Type, Union, get_origin

import httpx
import pydantic



URLType = Union[str, httpx.URL]
HttpStatusCode = httpx.codes
Response = httpx.Response

_re_path_param = re.compile(r'^[a-zA-Z0-9_-]+$')


class ApiBaseModel(pydantic.BaseModel):
    """Base class of all generated API models."""

    # Methods to improve IPython experience

    def __dir__(self):
        return [arg[0] for arg in self.__repr_args__()]

    def _repr_pretty_(self, p, cycle):
        if cycle:
            p.text(f"{self.__class__.__name__}(...)")
        else:
            p.begin_group(2, f"{self.__class__.__name__}(")
            first = True
            for name, value in self.__repr_args__():
                if not first:
                    p.text(',')
                    p.breakable(' ')
                first = False
                if name is not None:
                    p.text(f"{name}=")
                p.pretty(value)
            p.end_group(2, ')')


# This interface is only for type checking.
if TYPE_CHECKING:
    class GeneratedInterfaceBaseType:
        async def get(self, url: URLType, *, params: Optional[Mapping[str, Any]] = None) -> Response:
            ...

        async def post_body(self, url: URLType, *, content: Union[str, bytes, Iterable[bytes], AsyncIterable[bytes]], content_type: str = 'application/json', params: Optional[Mapping[str, Any]] = None) -> Response:
            ...

        async def post_form(self, url: URLType, *, data: Optional[Mapping[str, Any]] = None, params: Optional[Mapping[str, Any]] = None) -> Response:
            ...

        async def post(self, url: URLType) -> Response:
            ...

        async def put(self, url: URLType, *, params: Optional[Mapping[str, Any]] = None) -> Response:
            ...

        async def delete(self, url: URLType, *, params: Optional[Mapping[str, Any]] = None) -> Response:
            ...
else:
    class GeneratedInterfaceBaseType:
        pass


class GeneratedInterfaceBase(GeneratedInterfaceBaseType):
    """Generated Cherwell API interface"""

    def validate_path_param(self, value: str, type: Type = str):
        if not _re_path_param.match(value):
            raise ValueError(f"Invalid path parameter: {value!r}")
        type_origin = get_origin(type)
        if type_origin is Literal:
            if value.lower() not in [a.lower() for a in type.__args__]:
                raise ValueError(
                    f"Parameter mismatch with literal: {value} not in {type}")


    def parse_response(self, response: Response, rtype: Type) -> Any:
        # If the response type has a `httpStatusCode` field, set it to the actual HTTP status code
        # and don't raise an exception if the HTTP request failed
        try:
            # issubclass can fail especially with types
            is_api = issubclass(rtype, ApiBaseModel)
        except:
            is_api = False

        if is_api:
            status_code = None
            try:
                # The Cherwell server sends the header: "Content-Type: application/json;charset=UTF-8"
                obj = rtype.parse_raw(response.content,
                                      content_type=response.headers.get(
                                          'content-type').split(';', 1)[0],
                                      encoding=response.encoding or 'utf-8')
                if hasattr(obj, 'httpStatusCode'):
                    obj.httpStatusCode = response.status_code  # type: ignore
                    status_code = response.status_code
            except pydantic.ValidationError:
                # Maybe the pydantic validation failed because the HTTP request failed
                if not response.is_success:
                    response.raise_for_status()
                raise
            # HTTP error: if we can respond with a httpStatusCode in the response, do so, otherwise raise an exception
            if not response.is_success:
                if status_code:
                    return obj
                response.raise_for_status()
            # HTTP success and pydantic validation succeeded
            return obj

        # Reach here if the type isn't a pydantic model
        if not response.is_success:
            response.raise_for_status()

        if rtype is str:
            return response.text
        if rtype is bytes:
            return response.content
        if rtype is None:
            return None

        return pydantic.parse_raw_as(rtype,
                                     response.content,
                                     content_type=response.headers.get(
                                         'content-type').split(';', 1)[0],
                                     encoding=response.encoding or 'utf-8')




import re
from functools import cached_property
from typing import Optional, Pattern, Union, cast

from async_lru import alru_cache

from cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject import (
    SchemaResponse,
    Summary,
)
from cherwell_pydantic_api.bo.modelgen.model_generator import PydanticModelGenerator
from cherwell_pydantic_api.bo.modelgen.repo import FileGenerator, ModelRepo
from cherwell_pydantic_api.bo.valid_schema import ValidSchema
from cherwell_pydantic_api.instance import Instance
from cherwell_pydantic_api.types import BusinessObjectType, BusObID
from cherwell_pydantic_api.utils import docwraps



try:
    import click
except ImportError:
    click = None


FilterType = Union[Pattern, str, None]


def compile_filter(filter: FilterType) -> Optional[Pattern]:
    if filter is None:
        return None
    if isinstance(filter, str):
        if filter == '':
            return None
        return re.compile(filter)
    return filter


class CollectorItem:
    "Contains a Business Object summary, and a verdict on whether to generate a model for it."

    def __init__(self,
                 bo_type: BusinessObjectType,
                 summary: Summary,
                 verdict: Optional[bool] = None,
                 schema: Optional[SchemaResponse] = None,
                 parent_item: Optional["CollectorItem"] = None):
        assert summary.busObId is not None
        self.summary = summary
        self.verdict = verdict
        self.bo_type = bo_type
        self.schema = schema
        self._children: set["CollectorItem"] = set()
        self.parent_item = parent_item
        if parent_item is not None:
            parent_item._children.add(self)

    @property
    def busobid(self) -> BusObID:
        return self.summary.busObId  # type: ignore

    @property
    def name(self) -> str:
        return self.summary.name  # type: ignore

    @cached_property
    def valid_schema(self) -> Optional[ValidSchema]:
        if self.schema is None:
            return None
        return ValidSchema.from_schema_response(self.schema)


    def to_dict(self) -> dict:
        "Convert to dictionary e.g. for DataFrame use"
        return {'name': self.name,
                'busobid': self.busobid,
                'verdict': self.verdict,
                'bo_type': self.bo_type,
                'num_fields': len(self.schema.fieldDefinitions) if (self.schema and self.schema.fieldDefinitions) else None,
                'displayName': self.summary.displayName,
                'lookup': self.summary.lookup,
                'major': self.summary.major,
                'supporting': self.summary.supporting,
                'parent_name': self.parent_item.name if self.parent_item else None,
                '__self': self,
                }


class Collector:
    "Collects Business Object summaries and schemas from a Cherwell instance, in order to filter them and generate models."

    def __init__(self, instance: Instance, *, bo_include_filter: FilterType = None, bo_exclude_filter: FilterType = None, verbose: bool = False):
        self._instance = instance
        self._bo_include_filter = compile_filter(bo_include_filter)
        self._bo_exclude_filter = compile_filter(bo_exclude_filter)
        self.items: list[CollectorItem] = []
        self.verbose = verbose


    @property
    def bo_include_filter(self) -> Optional[Pattern]:
        return self._bo_include_filter

    @bo_include_filter.setter
    def bo_include_filter(self, value: FilterType):
        self._bo_include_filter = compile_filter(value)

    @property
    def bo_exclude_filter(self) -> Optional[Pattern]:
        return self._bo_exclude_filter

    @bo_exclude_filter.setter
    def bo_exclude_filter(self, value: FilterType):
        self._bo_exclude_filter = compile_filter(value)


    def filter_bo(self, type: str, summary: Summary) -> bool:
        if self._bo_exclude_filter is not None:
            if self._bo_exclude_filter.match(summary.name):
                return False
        if self._bo_include_filter is not None:
            if self._bo_include_filter.match(summary.name):
                return True
        return False


    def verbose_report(self, item: CollectorItem):
        icon = "[**]" if item.verdict else "[  ]"
        if item.schema:
            fields = len(
                item.schema.fieldDefinitions) if item.schema.fieldDefinitions else 0
            rels = len(
                item.schema.relationships) if item.schema.relationships else 0
        else:
            fields = rels = '?'
        report = f" {item.bo_type[:3]:3} {item.name:30} {fields:>4} {rels:>4}  {item.busobid:24}"
        if click:
            click.secho(icon, nl=False,
                        bg='green' if item.verdict else 'red', fg='white')
            click.secho(report, fg='cyan')
        else:
            print(icon, report)


    @alru_cache
    @docwraps(Instance.get_bo_summaries)
    async def get_bo_summaries(self, bo_type: BusinessObjectType) -> list[Summary]:
        return await self._instance.get_bo_summaries(bo_type)


    @alru_cache(maxsize=1024)
    @docwraps(Instance.get_bo_schema)
    async def get_bo_schema(self, busobid: BusObID) -> Optional[SchemaResponse]:
        return await self._instance.get_bo_schema(busobid=busobid)


    async def select(self) -> list[CollectorItem]:
        "Get all the business object summaries from Cherwell and filter them."
        items = []
        for bo_type in cast(list[BusinessObjectType], ['Major', 'Supporting', 'Lookup', 'Groups']):
            for bo_summary in await self.get_bo_summaries(bo_type):
                verdict = self.filter_bo(bo_type, bo_summary)
                item = CollectorItem(
                    bo_type=bo_type, summary=bo_summary, verdict=verdict)
                items.append(item)
                # Add group members. There is only one level of groups according to Cherwell docs.
                if bo_summary.groupSummaries:
                    for group_summary in bo_summary.groupSummaries:
                        verdict = self.filter_bo(bo_type, group_summary)
                        sub_item = CollectorItem(
                            bo_type=bo_type, summary=group_summary, verdict=verdict, parent_item=item)
                        items.append(sub_item)
        return items


    async def collect(self) -> list[CollectorItem]:
        items = await self.select()
        for item in items:
            if item.verdict:
                schema = await self.get_bo_schema(item.summary.busObId)
            else:
                schema = None
            item.schema = schema
            if self.verbose:
                self.verbose_report(item)
        self.items = items
        return items


    def generate_models(self) -> FileGenerator:
        yield ('__init__.py', '### autogenerated by cherwell_pydantic_api ###\n')
        model_generator = PydanticModelGenerator(self._instance.settings)
        yield ('__base.py', model_generator.generate_base())
        for item in self.items:
            schema = self._instance.bo.get_schema(item.busobid)
            yield (f"{schema.identifier}.py", model_generator.generate_model(schema))


    def save_models(self, repo: ModelRepo):
        return repo.save(instance=self._instance, file_type='model', file_generator=self.generate_models())


    def clear_caches(self):
        self.get_bo_summaries.cache_clear()
        self.get_bo_schema.cache_clear()


    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self._instance}, bo_include_filter={self._bo_include_filter}, bo_exclude_filter={self._bo_exclude_filter})"




from collections import defaultdict
from typing import Iterable, Mapping, Optional, Union
from weakref import WeakValueDictionary

from pydantic import AnyHttpUrl, BaseModel

from cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject import (
    Relationship,
    SchemaResponse,
    Summary,
)
from cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.Core import ServiceInfoResponse
from cherwell_pydantic_api.bo.valid_schema import ValidSchema
from cherwell_pydantic_api.bo.wrapper import BusinessObjectWrapper, BusinessObjectWrapperBase
from cherwell_pydantic_api.interfaces import ApiRequesterInterface
from cherwell_pydantic_api.types import BusObID, BusObIdentifier, RelationshipID



# Registry of business object schemas.
# It appears that the API returns schema fields in the same order, but not the relationships.



class ServiceInfoModel(BaseModel):
    apiVersion: Optional[str] = None
    csmCulture: Optional[str] = None
    csmVersion: Optional[str] = None
    base_url: Optional[AnyHttpUrl] = None


class BusinessObjectModelRegistryMixin:
    pass


class BusinessObjectRegistry(Mapping[str, BusinessObjectWrapperBase]):
    # There is exactly one BusinessObjectRegistry per Instance.
    # There can be more than one wrapper per BO but only one schema or summary.

    def __init__(self, instance: ApiRequesterInterface, wrapper_class: type[BusinessObjectWrapperBase] = BusinessObjectWrapper):
        self._instance = instance
        self._wrapper_class = wrapper_class
        self._schemas: dict[BusObID, ValidSchema] = {}
        self._summaries: dict[BusObID, Summary] = {}
        self._name_to_id: dict[BusObIdentifier, BusObID] = {}
        self._relationships: dict[RelationshipID, Relationship] = {}
        self._bo_rels: defaultdict[BusObID,
                                   set[RelationshipID]] = defaultdict(set)
        self._wrappers: WeakValueDictionary[BusObID,
                                            wrapper_class] = WeakValueDictionary()
        self._service_info: Optional[ServiceInfoResponse] = None
        self._models: WeakValueDictionary[BusObID,
                                          type[BusinessObjectModelRegistryMixin]] = WeakValueDictionary()


    def marshal(self, include_summaries: bool = False) -> Iterable[tuple[str, str]]:
        yield ('instance_name.txt', self._instance.settings.name)
        for busobid, schema in self._schemas.items():
            yield (f'bo.{busobid}.json', schema.json(indent=2, exclude={'relationships'}, sort_keys=True, exclude_unset=True, exclude_defaults=True))
        # TODO: include relationships
        if include_summaries:
            for busobid, summary in self._summaries.items():
                if busobid in self._schemas:
                    continue
                yield (f'bo_sum.{busobid}.json', summary.json(indent=2, sort_keys=True))
        names_csv = [
            f"{name};{self._name_to_id[name]}\n" for name in self._name_to_id.keys()]
        names_csv.sort()
        yield ('names.csv', ''.join(names_csv))
        if self._service_info:
            si_model = ServiceInfoModel(
                apiVersion=self._service_info.apiVersion,
                csmCulture=self._service_info.csmCulture,
                csmVersion=self._service_info.csmVersion,
                base_url=self._instance.settings.base_url,
            )
            yield ('service_info.json', si_model.json(indent=2, sort_keys=True))


    def get_schema(self, busobid: BusObID) -> ValidSchema:
        return self._schemas[BusObID(busobid)]


    def register(self, schema: SchemaResponse):
        if schema.busObId is None:
            raise ValueError('SchemaResponse.busObId is None')
        if schema.name is None:
            raise ValueError('SchemaResponse.name is None')
        busobid = BusObID(schema.busObId)
        valid_schema = ValidSchema.from_schema_response(schema)
        identifier = valid_schema.identifier
        if busobid in self._schemas:
            if self._schemas[busobid] != valid_schema:
                raise ValueError(
                    f'SchemaResponse busObId {busobid} already registered and new schema is different')
            if identifier in self._name_to_id:
                if self._name_to_id[identifier] != busobid:
                    raise ValueError(
                        f'SchemaResponse identifier {identifier} already registered and new busObId is different')
        else:
            self._schemas[busobid] = valid_schema
            self._name_to_id[identifier] = busobid

        # Register relationships
        if schema.relationships:
            for rel in schema.relationships:
                assert rel.relationshipId is not None
                relid = RelationshipID(rel.relationshipId)
                assert relid in valid_schema.relationshipIds
                if relid in self._relationships:
                    if self._relationships[relid] != rel:
                        raise ValueError(
                            f'Relationship {relid} already registered and new relationship is different')
                    if not relid in self._bo_rels[busobid]:
                        # TODO: This occurs in group summaries
                        continue
                        raise ValueError(
                            f"Relationship {relid} is not in busObId {busobid}")
                else:
                    self._relationships[relid] = rel
                    self._bo_rels[busobid].add(relid)
                    # TODO: Make a ValidRelationship wrapper like ValidSchema
                    assert rel.target is not None
                    self._bo_rels[rel.target].add(relid) # type: ignore


    def register_summary(self, summary: Summary):
        # TODO: decide if this is needed, if so make it more useful
        if summary.busObId is None:
            raise ValueError('Summary.busObId is None')
        if summary.name is None:
            raise ValueError('Summary.name is None')
        busobid = BusObID(summary.busObId)
        identifier = BusObIdentifier(summary.name)
        if identifier in self._name_to_id:
            if self._name_to_id[identifier] != busobid:
                raise ValueError(
                    f'Summary name {summary.name} already registered and new busObId is different')
        else:
            self._name_to_id[identifier] = busobid
        if busobid in self._summaries:
            if self._summaries[busobid] != summary:
                raise ValueError(
                    f'Summary busObId {summary.busObId} already registered and new summary is different')
            return
        self._summaries[busobid] = summary


    def register_model(self, busobid: BusObID, model: type[BusinessObjectModelRegistryMixin]):
        self._models[busobid] = model


    def get_model(self, busobid: BusObID) -> type[BusinessObjectModelRegistryMixin]:
        return self._models[busobid]


    def register_service_info(self, service_info: ServiceInfoResponse):
        self._service_info = service_info


    def set_wrapper_class(self, wrapper_class: type[BusinessObjectWrapperBase]):
        if wrapper_class != self._wrapper_class:
            self._wrappers.clear()
        self._wrapper_class = wrapper_class


    def get_wrapper(self, busobid: BusObID) -> BusinessObjectWrapperBase:
        if busobid not in self._wrappers:
            wrapper = self._wrapper_class(
                schema=self._schemas[busobid], instance=self._instance)
            self._wrappers[busobid] = wrapper
            return wrapper
        return self._wrappers[busobid]


    def __getitem__(self, item: Union[BusObID, BusObIdentifier, str]) -> BusinessObjectWrapperBase:
        """Look up a business object wrapper by identifer or ID."""
        try:
            busobid = BusObID(item)
        except ValueError:
            busobid = False
        if not busobid or busobid not in self._schemas:
            busobname = BusObIdentifier(item)
            busobid = self._name_to_id[busobname]
        return self.get_wrapper(busobid)


    def __iter__(self) -> Iterable[str]:
        """Iterate over the names of the registered business objects. Don't include summaries. (Not the business object IDs)"""
        return iter([k for k, v in self._name_to_id.items() if v in self._schemas])


    def __len__(self) -> int:
        """Return the number of registered business objects. Don't include summaries."""
        return len(self._schemas)


    def __getattr__(self, attr: Union[BusObIdentifier, str]) -> BusinessObjectWrapperBase:
        try:
            return self.__getitem__(attr)
        except KeyError:
            raise AttributeError(attr)


    def __repr__(self) -> str:
        return f'<{self.__class__.__name__}[{len(self._schemas)}] (instance={self._instance.settings.name})>'


    def __dir__(self) -> Iterable[str]:
        return [k for k in self.__iter__()] + [k for k in super().__dir__()]



import datetime
from decimal import Decimal
from typing import Optional



def validator_datetime(value: str) -> Optional[datetime.datetime]:
    if not value:
        return None
    value = value.replace('/', '.')
    try:
        return datetime.datetime.strptime(value, '%d.%m.%Y %H:%M')
    except:
        pass
    try:
        return datetime.datetime.strptime(value, '%d.%m.%Y %H:%M:%S')
    except:
        pass
    raise ValueError(f"Invalid datetime: {value!r}")


def validator_date(value: str) -> Optional[datetime.date]:
    if not value:
        return None
    value = value.replace('/', '.')
    return datetime.datetime.strptime(value, '%d.%m.%Y').date()


def validator_time(value: str) -> Optional[datetime.time]:
    # TODO: better handling of invalid values
    if not value:
        return None
    try:
        return datetime.datetime.strptime(value, '%H:%M').time()
    except:
        pass
    try:
        return datetime.datetime.strptime(value, '%H:%M:%S').time()
    except:
        pass
    return None


def validator_int(value: str) -> int:
    # TODO: better handling of invalid values
    if '.' in value:
        value = value.split('.')[0]
    return int(value.translate(str.maketrans("',", "__")))


def validator_decimal(value: str) -> Decimal:
    return Decimal(value.translate(str.maketrans("',", "__")))



from typing import Optional, cast

from pydantic import Field, constr

from cherwell_pydantic_api._generated.api.models.Trebuchet.WebApi.DataContracts.BusinessObject import (
    FieldDefinition,
    GridDefinition,
    Relationship,
    SchemaResponse,
)
from cherwell_pydantic_api.generated_api_utils import ApiBaseModel
from cherwell_pydantic_api.types import BusObID, BusObIdentifier, FieldID, FieldIdentifier, RelationshipID
from cherwell_pydantic_api.utils import fieldid_parts



class ValidFieldDefinition(FieldDefinition):
    fieldId: FieldID
    name: str
    type: str
    short_field_id: FieldID = Field(description="Just the field ID, without 'BO:..,FI:...'")
    identifier: FieldIdentifier

    def __init__(self, **data):
        data['short_field_id'] = fieldid_parts(data['fieldId'])['FI']
        data['identifier'] = FieldIdentifier(data['name'])
        super().__init__(**data)

    def to_dict(self) -> dict:
        "Convert to dictionary e.g. for DataFrame use"
        return {'name': self.name,
                'type': self.type,
                'short_field_id': self.short_field_id,
                'identifier': self.identifier,
                '__self': self,
                }



class ValidSchema(ApiBaseModel):
    "Corresponds to the type SchemaResponse, but with stricter types and omitting relationship details"
    busObId: BusObID
    fieldDefinitions: list[ValidFieldDefinition]
    firstRecIdField: Optional[str] = None
    gridDefinitions: Optional[list[GridDefinition]] = None
    name: str
    recIdFields: Optional[str] = None
    stateFieldId: Optional[str] = None
    states: Optional[str] = None
    identifier: BusObIdentifier = Field(description="ASCII lowercase of name, to be used as python identifier")
    relationshipIds: list[RelationshipID]

    @classmethod
    def from_schema_response(cls, schema: SchemaResponse):
        if schema.hasError or not schema.busObId or not schema.name or not schema.fieldDefinitions:
            raise ValueError(f'SchemaResponse not valid: {schema}')
        if schema.relationships:
            relIds = [RelationshipID(r.relationshipId) for r in schema.relationships if r.relationshipId]
            relIds.sort()
        else:
            relIds = []
        return cls(busObId=BusObID(schema.busObId),
                   fieldDefinitions=[ValidFieldDefinition(**field.dict()) for field in schema.fieldDefinitions],
                   firstRecIdField=schema.firstRecIdField,
                   gridDefinitions=schema.gridDefinitions,
                   name=schema.name,
                   identifier=BusObIdentifier(schema.name),
                   recIdFields=schema.recIdFields,
                   relationshipIds=relIds,
                   stateFieldId=schema.stateFieldId,
                   states=schema.states,)



#!/usr/bin/env python

import sys
from pathlib import Path



if sys.version_info < (3, 9):
    sys.stderr.write("Python 3.9 or higher is required.")
    sys.exit(1)
try:
    import cherwell_pydantic_api
    from cherwell_pydantic_api.settings import Settings
except ImportError as e:
    sys.stderr.write(
        "cherwell_pydantic_api is not installed properly as a package. Please check the documentation.")
    sys.stderr.write(f"Error: {e}")
    sys.exit(1)
try:
    import click
except ImportError:
    sys.stderr.write("""To use the CLI, click must be installed.
One way of achieving this is to install cherwell_pydantic_api with the modelgen option.
But you can also install click separately.

To fix this, run pip or poetry as follows:
    pip install cherwell_pydantic_api[modelgen]
or
    poetry add -E modelgen cherwell_pydantic_api

""")
    sys.exit(1)


def check_envpath(offer_setup=True):
    envpath = Path.cwd().joinpath('cherwell.env')
    if not envpath.exists():
        click.secho(
            f"\nNo cherwell.env file found in the current path, {Path.cwd()}.\nYou can now create one, or press Ctrl+C to abort.\n", fg='yellow')
        from .welcome import initial_setup
        initial_setup(envpath)
    return envpath


def welcome_banner():
    click.echo()
    lines = ['Welcome to cherwell_pydantic_api!', '', 'For more information, please visit:', '', 'https://github.com/dataway/cherwell_pydantic_api']
    width = max([len(line) for line in lines])
    hash = '\u2592'
    style = {'fg': 'white', 'bg': 'blue'}
    click.secho(hash * (width + 6), **style)
    for line in lines:
        click.secho(f"{hash}  {line}{' ' * (width - len(line))}  {hash}", **style)
    click.secho(hash * (width + 6), **style)
    click.echo()



def cli_welcome():
    "Check for existing env file and offer to create one if not found. Return 0 if successful, 1 if aborted."
    try:
        settings = Settings()
    except:
        settings = None
    if settings is None or not settings.suppress_banner:
        welcome_banner()

    try:
        envpath = check_envpath()
    except click.exceptions.Abort:
        click.secho("\nAborted.\n", fg='red')
        return 1
    if settings is None or not settings.suppress_banner:
        click.secho(f"\nFound existing cherwell.env file at {envpath}\n", fg='green')
    return 0


@click.group()
def root():
    pass


def cli_normal():
    from . import check
    root.add_command(check.check)
    from . import ipython
    root.add_command(ipython.ipython_shell)
    root()


def cli():
    "Main CLI entry point."
    if len(sys.argv) == 1: # No CLI arguments
        r = cli_welcome()
        if r != 0:
            return r
    return cli_normal()


if __name__ == '__main__':
    cli()




# generated by datamodel-codegen:
#   filename:  csm_api-swagger.json

from __future__ import annotations

from enum import Enum
from typing import Dict, List, Optional

import cherwell_pydantic_api.types as ct
from cherwell_pydantic_api.generated_api_utils import ApiBaseModel, HttpStatusCode

from .... import CherwellObjectID
from . import Core


class ApprovalReadResponse(ApiBaseModel):
    properties: Optional[Dict[str, CherwellObjectID]] = None
    links: Optional[List[Core.CherwellLink]] = None
    errorCode: Optional[str] = None
    errorMessage: Optional[str] = None
    hasError: Optional[bool] = None
    httpStatusCode: Optional[HttpStatusCode] = None


import cherwell_pydantic_api.types as ct


class GetApprovalsResponse(ApiBaseModel):
    totalRecords: Optional[int] = None
    approvals: Optional[List[ApprovalReadResponse]] = None
    errorCode: Optional[str] = None
    errorMessage: Optional[str] = None
    hasError: Optional[bool] = None
    httpStatusCode: Optional[HttpStatusCode] = None



from functools import wraps



def async_command(command):
    @wraps(command)
    def wrapper(*args, **kwargs):
        import asyncio
        return asyncio.run(command(*args, **kwargs))
    return wrapper



import logging
from pathlib import Path
from typing import Iterable, Literal, Optional

from dulwich import porcelain
from dulwich.errors import NotGitRepository
from dulwich.repo import Repo

from cherwell_pydantic_api.instance import Instance
from cherwell_pydantic_api.settings import Settings



logger = logging.getLogger(__name__)


FileTypeLiteral = Literal['registry', 'model']
FileGenerator = Iterable[tuple[str, str]]


class ModelRepo:
    def __init__(self, create: bool = False):
        is_new = False
        self._repo_dir = Settings().repo_dir.absolute()
        if not self._repo_dir.exists():
            if create:
                self._repo_dir.mkdir(parents=True)
            else:
                raise ValueError(
                    f'Repo directory {self._repo_dir} does not exist')
        try:
            self._repo = Repo(root=str(self._repo_dir))
        except NotGitRepository:
            if create:
                self._repo = Repo.init(path=str(self._repo_dir))
                is_new = True
            else:
                raise
        if is_new:
            self._repo_dir.joinpath('.gitignore').write_bytes(
                Path(__file__).parent.joinpath('../templates/dot_gitignore').read_bytes())
            self._repo.stage('.gitignore')
            commit = f'cherwell_pydantic_api: initial repo setup'
            author = Settings().repo_author
            self._repo.do_commit(message=commit.encode(),
                                 author=author.encode())


    def prepare_save(self, *, instance_name: Optional[str] = None, instance: Optional[Instance] = None) -> tuple[Instance, Path]:
        # ensure repo is clean
        repo_status = porcelain.status(self._repo)  # type: ignore
        if repo_status.unstaged or repo_status.untracked or repo_status.staged['add'] or repo_status.staged['delete'] or repo_status.staged['modify']:
            raise ValueError(
                'Repo is not clean, please commit or stash your changes before saving')

        # resolve arguments
        if instance is None:
            if instance_name is None:
                raise ValueError(
                    'Either instance_name or instance must be given')
            instance = Instance._instances[instance_name]
        else:
            instance_name = instance.settings.name

        instance_dir = self._repo_dir / instance.settings.get_repo_subpackage()
        if not instance_dir.exists():
            logging.info(
                f'Creating instance directory {instance_dir} for instance {instance_name}')
            instance_dir.mkdir()
        return (instance, instance_dir)


    def save(self, *,
             instance_name: Optional[str] = None,
             instance: Optional[Instance] = None,
             file_type: FileTypeLiteral = 'registry',
             file_generator: FileGenerator,
             commit: bool = True):
        instance, instance_dir = self.prepare_save(
            instance_name=instance_name, instance=instance)
        file_type_dir = instance_dir / file_type
        if not file_type_dir.exists():
            logging.info(
                f'Creating {file_type} directory {file_type_dir} for instance {instance_name}')
            file_type_dir.mkdir()
        for filename, content in file_generator:
            filepath = file_type_dir / filename
            logging.info(f'Writing file {filepath}')
            filepath.write_text(content)
            self._repo.stage(str(filepath.relative_to(self._repo_dir)))
        if commit:
            self.commit(instance=instance, file_type=file_type)


    def commit(self, *, instance: Instance, file_type: FileTypeLiteral) -> Optional[bytes]:
        # Check if there are any changes to commit (because dulwich allows empty commits)
        repo_status = porcelain.status(self._repo)  # type: ignore
        if not repo_status.staged['add'] and not repo_status.staged['delete'] and not repo_status.staged['modify']:
            logging.info(
                f'No {file_type} changes to commit for instance {instance.settings.name}')
            return None
        commit = f'cherwell_pydantic_api: update {file_type} for instance {instance.settings.name}'
        author = Settings().repo_author
        branch = instance.settings.get_repo_branch()
        ref = f'refs/heads/{branch}'
        porcelain.update_head(self._repo, ref)  # this is like git checkout
        logging.info(
            f'Committing {file_type} to {branch} for instance {instance.settings.name}')
        return self._repo.do_commit(message=commit.encode(),
                                    author=author.encode(), ref=ref.encode())


    def save_all(self):
        for instance in Instance._instances.values():
            self.save(instance=instance, file_generator=instance.bo.marshal())



import keyword
import re
from typing import TYPE_CHECKING, Annotated, Any, AsyncIterable, Iterable, Literal, NewType, Optional, Union



if TYPE_CHECKING:
    IDType = NewType
    IdentifierType = NewType
    LCIdentifierType = NewType
else:
    def IDType(name, tp):
        """Create a subtype of str with a constructor that converts identifiers to lower case. Non-hex characters
        will throw a ValidError. For the type checker it is the same as NewType"""
        def id_type(x):
            x = x.lower()
            if not re.match(r'^[0-9a-f]+$', x):
                raise ValueError(f"Invalid {name}: {x}")
            return x
        id_type.__name__ = name
        id_type.__supertype__ = tp
        return id_type

    def IdentifierType(name, tp):
        """Create a subtype of str with a constructor that converts field names, etc. to a string
        that can be used as a Python identifier, except python keywords and fields that begin with a digit,
        in which case a prefix of I_ will be added (capitals). For the type checker it is the same as NewType"""
        def identifier_type(x):
            x = re.sub(r'[^0-9A-Za-z_]', '_', x)
            if x[0].isdigit() or keyword.iskeyword(x):
                x = 'I_' + x
            return x
        identifier_type.__name__ = name
        identifier_type.__supertype__ = tp
        return identifier_type

    def LCIdentifierType(name, tp):
        """Create a subtype of str with a constructor that converts field names, etc. to a _lowercase_ string
        that can be used as a Python identifier, except python keywords and fields that begin with a digit,
        in which case a prefix of I_ will be added (capitals). For the type checker it is the same as NewType"""
        def identifier_type(x):
            x = re.sub(r'[^0-9a-z_]', '_', x.lower())
            if x[0].isdigit() or keyword.iskeyword(x):
                x = 'I_' + x
            return x
        identifier_type.__name__ = name
        identifier_type.__supertype__ = tp
        return identifier_type


CherwellObjectID = dict[str, Any]
BusObID = IDType("BusObID", str)
BusObRecID = IDType("BusObRecID", str)
BusObIDParamType = BusObID
BusObIdentifier = LCIdentifierType("BusObIdentifier", str)
FieldID = IDType("FieldID", str)
FieldIdentifier = IdentifierType("FieldIdentifier", str)
RelationshipID = IDType("RelationshipID", str)
FileType = Union[str, bytes, Iterable[bytes], AsyncIterable[bytes]]
FileUpload = Annotated[FileType, "FileUpload"]
FileDownload = Annotated[FileType, "FileDownload"]
StringResponse = NewType("StringResponse", str)


BusinessObjectType = Literal["All", "Major", "Supporting", "Lookup", "Groups"]
RecordAttachmentType = Literal[
    "None",
    "File",
    "FileManagerFile",
    "BusOb",
    "History",
    "Other",
    "Solution",
    "UsedAsSolution",
    "ExternalSolution",
]
RecordAttachmentType.__doc__ = """Record attachment type:
        None - Not applicable to the REST API.
        File - Linked files.
        FileManagerFile - Imported files.
        BusOb - Attached Business Objects.
        History - Information about the attachment, if any is available. For example, an e-mail message may store the name of an attachment sent.
        Other - Not applicable to the REST API."""

AttachmentType = Literal["Imported", "Linked", "URL"]
AttachmentType.__doc__ = """For file types, select the type of attachment:
        Imported - Attachment was imported into database.
        Linked - Attachment is linked to an external file.
        URL - Attachment is a URL."""


class CherwellAPIError(Exception):
    def __init__(self, msg: str = "Cherwell API error", *, errorCode: Optional[str] = None,
                 errorMessage: Optional[str] = None,
                 httpStatusCode: Optional[Any] = None):
        message = f"{msg}: {errorCode=} {errorMessage=} {httpStatusCode=}"
        super().__init__(message)



from typing import Any, Callable
from cherwell_pydantic_api.types import FieldID



def fieldid_parts(field_id: str) -> dict[str, str]:
    """The fieldId has the format: 'BO:947dfef9ae6b072ca567a444dca79d9f9ea47a112f,FI:9492462c89a8f7785f1537412ca51a2f218293edb0,RE:9492535dfbf4286b49047f45bab2f201737e739a46'
    This method returns a dict with the keys 'BO', 'FI', 'RE', etc. and the values are the corresponding IDs."""
    r = {}
    for part in field_id.split(','):
        k, v = part.split(':', 1)
        assert k in ('BO', 'FI', 'RE')
        r[k.upper()] = FieldID(v)
    return r


def docwraps(func: Callable[..., Any]) -> Callable[..., Any]:
    """Decorator to copy the docstring from the wrapped function to the wrapper function."""
    def wrapper(wrapper_func: Callable[..., Any]) -> Callable[..., Any]:
        if wrapper_func.__doc__ is None:
            wrapper_func.__doc__ = f"Wraps {func.__qualname__}"
        else:
            wrapper_func.__doc__ += f"\nWraps {func.__qualname__}"
        if func.__doc__ is not None:
            wrapper_func.__doc__ += ":\n" + func.__doc__
        return wrapper_func
    return wrapper



from typing import Optional

import click

from ..instance import Instance
from .utils import async_command



@click.command()
@click.option('--instance-name', '-I', help='The name of the instance to use. If not specified, the default instance will be used.')
@async_command
async def check(instance_name: Optional[str] = None):
    "Check the connectivity to the Cherwell API."
    instance = Instance.use(instance_name)
    service_info = await instance.get_service_info()
    click.echo(f"OK: instance {instance.settings.name} at {instance.settings.base_url}")
    click.echo(service_info)
    await instance.authenticate()
    click.echo('Authentication OK')
    await instance.logout()
    click.echo('Logout OK')
